__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from contributor_ci.logger import logger
from contributor_ci.main.extractor import GitHubExtractorBase


class RepoUsers(GitHubExtractorBase):

    name = "repo_users"
    description = "extract repositories worked on for external and internal users."
    depends_on = ["users"]  # users already depends on repos
    filenames = ["internal-users", "external-users"]

    def extract(self):
        """
        Extract metadata about internal and exernal users.
        """
        self._data["internal-repos"] = {}
        self._data["external-repos"] = {}

        repo_query = self.get_local_query("repo-users.gql")

        # Load in dependency files
        repos = self.load_dependency_file("repos")
        internal = self.load_dependency_file("internal-users")
        # external = self.load_dependency_file("external-users")

        repolist = sorted(repos.data.keys())
        memberlist = sorted(internal.data.keys())

        for repo in repolist:
            repo_user, repo_name = repo.split("/")

            try:
                out = self.manager.queryGitHubFromFile(
                    repo_query,
                    {
                        "ownName": repo_user,
                        "repoName": repo_name,
                        "numUsers": 50,
                        "pgCursor": None,
                    },
                    paginate=True,
                    cursorVar="pgCursor",
                    keysToList=["data", "repository", "mentionableUsers", "nodes"],
                )
            except Exception as error:
                logger.warning("Warning: Could not complete '%s': %s" % (repo, error))
                continue

            # Update collective data
            for user in out["data"]["repository"]["mentionableUsers"]["nodes"]:
                username = user["login"]
                if username in memberlist:
                    self.add_contributions(username, "internal-repos", repo)
                else:
                    self.add_contributions(username, "external-repos", repo)

    def add_contributions(self, username, key, repo):
        """
        Given a username and a key, add a repository node
        """
        if username not in self._data[key]:
            self._data[key][username] = {}
        if "contributedLabRepositories" not in self._data[key][username]:
            self._data[key][username]["contributedLabRepositories"] = {"nodes": []}
        self._data[key][username]["contributedLabRepositories"]["nodes"].append(repo)
        self._data[key][username]["contributedLabRepositories"]["nodes"].sort()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"


import abc
import importlib
import inspect
import os
from collections import defaultdict
from collections.abc import Mapping
from datetime import datetime

from scraper.github import queryManager as qm

import contributor_ci.utils
from contributor_ci.logger import logger
from contributor_ci.main.settings import SettingsBase

here = os.path.abspath(os.path.dirname(__file__))


class ExtractorFinder(Mapping):
    """
    An ExtractorFinder keeps a cache of available extractors
    """

    _extractors = {}

    def __init__(self, save_format=None):
        """
        Instantiate an extractor
        """
        self.save_format = save_format
        self.collection_path = os.path.join(here, "extractors")
        self.update()

    def __getitem__(self, name):
        return self._extractors.get(name)

    def __iter__(self):
        return iter(self._extractors)

    def __len__(self):
        return len(self._extractors)

    def update(self):
        """
        Update extractors
        """
        self._extractors = self._load_extractors()

    def _load_extractors(self):
        """
        Load extractors based on listing folders in the collection.
        """
        lookup = {}
        for name in os.listdir(self.collection_path):
            extractor_dir = os.path.join(self.collection_path, name)
            extractor_file = os.path.join(extractor_dir, "extract.py")

            # Skip files in collection folder
            if os.path.isfile(extractor_dir):
                continue

            # Continue if the file doesn't exist
            if not os.path.exists(extractor_file):
                logger.debug(
                    "%s does not appear to have an extract.py, skipping."
                    % extractor_dir
                )
                continue

            # The class name means we split by underscore, capitalize, and join
            class_name = "".join([x.capitalize() for x in name.split("_")])
            module = "contributor_ci.main.extractors.%s.extract" % name
            lookup[name] = getattr(importlib.import_module(module), class_name)(
                self.save_format
            )
        return lookup


class ExtractorBase:

    name = "metric"
    description = "An abstract base extractor."
    date_time_format = "%Y-%m-%dT%H:%M:%S%z"
    filenames = []
    depends_on = []

    def __init__(self, save_format=None):

        # Empty settings to be replaced by loaded settings
        self.settings = SettingsBase()
        self._data = {}
        self.save_format = save_format or "year/month/day"

        # If we change the output format, we cannot reliably date the output
        self.force_extract = self.save_format != "year/month/day"

        # This should be over-ridden typically by client get_extractor
        self.outdir = os.path.abspath(".cci")

    @abc.abstractmethod
    def extract(self):
        """Run the extraction."""
        pass

    @property
    def classpath(self):
        return os.path.dirname(inspect.getfile(self.__class__))

    def require_github(self):
        """
        Require GitHub token in the environment
        """
        self.token = os.environ.get("GITHUB_TOKEN")
        if not self.token:
            self.token = os.environ.get("CCI_GITHUB_TOKEN")
        if not self.token:
            logger.exit(
                "GITHUB_TOKEN or CCI_GITHUB_TOKEN is required to use %s" % self.name
            )

    def get_results(self):
        return self._data

    def exists(self):
        """
        Given output files known to the extactor, determine if already run.
        """
        outdir = self.get_dated_outdir()
        exists = True
        for name in self.filenames:
            outfile = "cci-%s.json" % name
            outfile = os.path.join(outdir, outfile)
            if os.path.exists(outfile) and self.force_extract:
                logger.debug(
                    "%s exists, but not reliably from today! Needs run." % outfile
                )
                exists = False
            elif not os.path.exists(outfile):
                logger.debug("%s does not exist, needs run." % outfile)
                exists = False
        return exists

    def get_dependency_file(self, name):
        """
        Given the name of a dependency file, retrieve it for the current data.
        """
        outdir = self.get_dated_outdir()
        filename = os.path.join(outdir, "cci-%s.json" % name)
        if not os.path.exists(filename):
            logger.exit("Dependency file %s does not exist." % filename)
        return filename

    def get_recent_dependency_file(self, name):
        """
        Given the name of a dependency file, find the latest.
        """
        filename = os.path.join(self.outdir, "data", "latest", "cci-%s.json" % name)
        if os.path.exists(filename):
            return filename

    def get_local_query(self, path):
        """
        Given the path of a graph ql (.gql) file in the collection, return it
        """
        path = os.path.join(self.classpath, path)
        if not os.path.exists(path):
            logger.exit(
                "Required GraphQL file %s does not exist alongside collection." % path
            )
        return path

    def get_dated_outdir(self):
        """
        Return nested directory name with year, month, day within data
        """
        outdir = self.outdir
        now = datetime.now()

        # Organize outdir by default year/month/day or custom
        outdir = os.path.join(outdir, "data")
        for level in self.save_format.split("/"):
            if level.lower() == "year":
                outdir = os.path.join(outdir, str(now.year))
            elif level.lower() == "month":
                outdir = os.path.join(outdir, str(now.month))
            elif level.lower() == "day":
                outdir = os.path.join(outdir, str(now.day))
        return outdir

    def save_json(self):
        """
        Save results to json, organized by type and date.
        """
        results = self.get_results()
        outdir = self.get_dated_outdir()

        # Also save as latest results
        latest = os.path.join(self.outdir, "data", "latest")

        for dirname in [latest, outdir]:
            if not os.path.exists(dirname):
                os.makedirs(dirname)

        for name, result in results.items():
            outname = "cci-%s.json" % name
            if result:
                for dirname in [outdir, latest]:
                    outfile = os.path.join(dirname, outname)
                    contributor_ci.utils.write_json(result, outfile)


class GitHubExtractorBase(ExtractorBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Requires export of GITHUB_TOKEN
        self._manager = None

    def load_dependency_file(self, name):
        """
        Given the name of a dependency, find and load the file
        """
        filename = self.get_dependency_file(name)
        return qm.DataManager(filename, True)

    def load_recent_dependency_file(self, name):
        """
        Given the name of a dependency, find and load the file
        """
        filename = self.get_recent_dependency_file(name)
        if filename:
            return qm.DataManager(filename, True)

    @property
    def manager(self):
        """
        Get or initialize the GitHub GraphQL Manager
        """
        self.require_github()
        if not self._manager:
            # Create a query manager
            self._manager = qm.GitHubQueryManager(self.token)
        return self._manager


class ExtractorResolver:
    """
    Given an extractor, create a graph (DAG) to determine execution order.
    """

    def __init__(self):
        self.graph = defaultdict(list)  # dictionary containing adjacency List

    @property
    def vertices(self):
        return len(self.graph)

    def add_extractor(self, ext):
        """
        Add an extractor to the graph, regardless of dependencies
        """
        self.graph[ext] = []

    def add_dependency(self, ext, depends_on):
        """
        Add a dependency to the graph
        """
        self.graph[ext].append(depends_on)

    @property
    def keys(self):
        return list(self.graph)

    def order_helper(self, v, seen, result):
        """
        A helper function to order the dependencies
        """
        # Lookup key to the index
        key = self.keys[v]

        # The current index is visited
        seen[v] = True

        # Visit unvisited extractors
        for i, ikey in enumerate(self.graph[key]):
            if not seen[i]:
                self.order_helper(i, seen, result)

        result.append(key)

    def resolve(self):
        """
        Resolve dependencies
        """
        seen = [False] * self.vertices
        result = []

        for i in range(self.vertices):
            if not seen[i]:
                self.order_helper(i, seen, result)

        # Return reversed
        return result[::-1]



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"


import sys

from contributor_ci.logger import logger
from contributor_ci.main import Client


def main(args, parser, extra, subparser):

    cli = Client(quiet=args.quiet, config_file=args.config_file, outdir=args.outdir)

    # If nothing provided, show help
    if not args.params:
        print(subparser.format_help())
        sys.exit(0)

    # The first "param" is either sort, edit, set of get
    command = args.params.pop(0)

    # For each new setting, update and save!
    if command == "sort":
        cli.settings.sort()
        cli.settings.save()

    # Add is used for a list
    elif command in ["add", "remove"]:
        key = args.params.pop(0)
        if not cli.settings.get(key):
            logger.exit("The key %s must exist to add to it." % key)

        if not isinstance(cli.settings.get(key), list):
            logger.exit("You can only add a new entry to a list.")

        for param in args.params:
            if command == "add":
                cli.settings.add(key, param)
            else:
                cli.settings.remove(key, param)
        cli.settings.sort()
        cli.settings.save()

    # For each get request, print the param pair
    elif command == "edit":
        cli.settings.edit()
    else:
        logger.error("%s is not a recognized command." % command)



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from contributor_ci.logger import logger
from contributor_ci.main.extractor import GitHubExtractorBase


class Topics(GitHubExtractorBase):

    name = "topics"
    description = "extract repository topics."
    depends_on = ["repos"]
    filenames = ["topics"]

    def extract(self):
        """
        Extract metadata about repository topics.
        """
        self._data[self.name] = {}
        topics_query = self.get_local_query("repo-topics.gql")

        # Load in dependency files
        repos = self.load_dependency_file("repos")

        repolist = sorted(repos.data.keys())

        for repo in repolist:
            logger.info("Looking up topics for %s" % repo)
            repo_user, repo_name = repo.split("/")

            try:
                out = self.manager.queryGitHubFromFile(
                    topics_query,
                    {
                        "ownName": repo_user,
                        "repoName": repo_name,
                        "numTopics": 25,
                        "pgCursor": None,
                    },
                    paginate=True,
                    cursorVar="pgCursor",
                    keysToList=["data", "repository", "repositoryTopics", "nodes"],
                )
            except Exception as error:
                logger.warning("Warning: Could not complete '%s': %s" % (repo, error))
                continue

            self._data[self.name][repo] = out["data"]["repository"]



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from datetime import date, timedelta

from contributor_ci.logger import logger
from contributor_ci.main.extractor import GitHubExtractorBase


def next_weekday(d, weekday):
    days_ahead = weekday - d.weekday()
    if days_ahead <= 0:
        days_ahead += 7
    return d + timedelta(days_ahead)


def toDate(isoStr):
    return next_weekday(date.fromisoformat(isoStr["starredAt"].split("T")[0]), 0)


class Stars(GitHubExtractorBase):

    name = "stars"
    description = "extract repository stars."
    depends_on = ["repos"]
    filenames = ["stars"]

    def extract(self):
        """
        Extract metadata about repository stars
        """
        self._data[self.name] = {}

        # Load in dependency files
        repos = self.load_dependency_file("repos")
        repolist = sorted(repos.data.keys())

        # load the query
        star_query = self.get_local_query("repo-stargazers.gql")

        for i, repo in enumerate(repolist):
            logger.info(
                "Parsing stargazer history for repository %s, %s of %s"
                % (repo, i, len(repolist))
            )
            repo_user, repo_name = repo.split("/")

            try:
                out = self.manager.queryGitHubFromFile(
                    star_query,
                    {
                        "ownName": repo_user,
                        "repoName": repo_name,
                        "numUsers": 100,
                        "pgCursor": None,
                    },
                    paginate=True,
                    cursorVar="pgCursor",
                    keysToList=["data", "repository", "stargazers", "edges"],
                )
            except Exception as error:
                logger.warning("Warning: Could not complete '%s': %s" % (repo, error))
                continue

            self._data[self.name][repo] = out["data"]["repository"]

        for repo in self._data[self.name]:
            dateRange = list(
                map(toDate, self._data[self.name][repo]["stargazers"]["edges"])
            )
            dateList = []
            dateElement = {"date": None, "value": None}
            for dateEntry in dateRange:
                if not dateElement["date"]:
                    dateElement["date"] = dateEntry.isoformat()
                    dateElement["value"] = 1
                elif dateElement["date"] == dateEntry.isoformat():
                    dateElement["value"] += 1
                else:
                    dateList.append(dateElement.copy())
                    dateElement["date"] = dateEntry.isoformat()
                    dateElement["value"] = 1
            self._data[self.name][repo] = dateList



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"

import re

from contributor_ci.logger import logger
from contributor_ci.main.extractor import GitHubExtractorBase


class CreationHistory(GitHubExtractorBase):

    name = "creation_history"
    description = "extract creation history for repositories."
    depends_on = ["repos"]
    filenames = ["creation_history"]

    def extract(self):
        """
        Extract metadata about creation history
        """
        self._data[self.name] = {}

        # Load in dependency files
        repos = self.load_dependency_file("repos")
        repolist = sorted(repos.data.keys())

        query_commits = "/repos/OWNNAME/REPONAME/commits?until=CREATETIME&per_page=100"
        query_commits_in = "/repos/OWNNAME/REPONAME/commits?per_page=100"

        # A recent dependency file is not a requirement
        history = self.load_recent_dependency_file("creation_history")
        if history:
            self._data[self.name] = history.data

        for i, repo in enumerate(repolist):
            logger.info(
                "Parsing creation history for repository %s, %s of %s"
                % (repo, i, len(repolist))
            )
            repo_user, repo_name = repo.split("/")

            # History doesn't change, only update new repos or those that had no previous commits
            if repo in self._data[self.name]:
                if self._data[self.name][repo].get("FirstCommitAt"):
                    logger.info("Already recorded data for '%s'" % (repo))
                    continue

            # Repository creation date
            repo_data = {"createdAt": repos.data[repo]["createdAt"]}

            # Commit timestamps
            gitquery2 = re.sub("OWNNAME", repo_user, query_commits)
            gitquery2 = re.sub("REPONAME", repo_name, gitquery2)
            gitquery2 = re.sub("CREATETIME", repo_data["createdAt"], gitquery2)

            try:
                out = self.manager.queryGitHub(gitquery2, rest=True, paginate=True)
            except Exception as error:
                out = []
                logger.warning("Could not complete '%s':%s" % (repo, error))

            # Update repo data
            repo_data["commitTimestamps"] = []
            self.update_commits(out, repo_data, repo)

            # If no pre-GitHub commits, check the greater commit history
            if repo_data["commitTimestamps"] and repo_data["commitTimestamps"][0]:
                repo_data["initBeforeGitHubRepo"] = True
            else:
                repo_data["initBeforeGitHubRepo"] = False

            gitquery3 = re.sub("OWNNAME", repo_user, query_commits_in)
            gitquery3 = re.sub("REPONAME", repo_name, gitquery3)

            try:
                out = self.manager.queryGitHub(gitquery3, rest=True, paginate=True)
            except Exception as error:
                print("Warning: Could not complete '%s': %s" % (repo, error))

            # Sort and save earlist commit date
            self.update_commits(out, repo_data, repo)
            repo_data["commitTimestamps"].sort()

            firstdate = None
            if repo_data["commitTimestamps"]:
                firstdate = repo_data["commitTimestamps"][0]
            repo_data["firstCommitAt"] = firstdate

            # We only need the first commit
            del repo_data["commitTimestamps"]

            self._data[self.name][repo] = repo_data

        # Clean up repos no longer in list
        updated = {}
        for repo in self._data[self.name]:
            if repo in repolist:
                updated[repo] = self._data[self.name][repo]

        self._data[self.name] = updated

    def update_commits(self, commits, repo_data, repo):
        """
        Update list of commit timestamps with new commit dates.
        """
        try:
            for commit in commits:
                repo_data["commitTimestamps"].append(
                    commit["commit"]["committer"]["date"]
                )
        except NameError:
            logger.info("Could not get pre-GitHub commits for '%s'" % (repo))
        return repo_data



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"

import os

import contributor_ci.utils as utils

# Replacements can currently be made for the database_file and lmod_base
install_dir = utils.get_installdir()
reps = {"$install_dir": install_dir, "$root_dir": os.path.dirname(install_dir)}

# The default settings file in the install root
default_settings_file = os.path.join(os.getcwd(), "contributor-ci.yaml")



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2020-2021, Vanessa Sochat"
__license__ = "MPL 2.0"

import os
import re

from contributor_ci.logger import logger

from .terminal import run_command


class GitManager:
    """
    Interact with a Git repository.
    """

    def __init__(self, folder=None, quiet=False):
        """
        Initialize a git manager.

        The folder can be empty if intending to
        init a new repository, or not existing if a clone is intended.
        """
        self.folder = folder or ""
        self.quiet = quiet
        self.repo = None
        self.update_repo(self.folder)

    def update_repo(self, dest):
        self.git_dir = os.path.join(dest, ".git")
        if os.path.exists(self.git_dir):
            self.repo = dest
            self.folder = dest

    @property
    def reponame(self):
        """
        Get the respository name.
        """
        if self.repo:
            name = self.run_command(
                self.init_cmd(self.repo) + ["config", "--get", "remote.origin.url"]
            ).strip()
            return (
                re.sub("http://|https://|[.]git|github.com|git@", "", name)
                .strip("/")
                .strip(":")
            )

    @property
    def reponame_flat(self):
        return self.reponame.replace("/", "-")

    def add(self, filename=".", dest=None):
        """
        Add a file to the git repository
        """
        dest = dest or self.folder or ""
        return self.run_command(self.init_cmd(dest) + ["add", filename])

    def commit(self, message, dest=None):
        """
        Commit to a particular directory
        """
        dest = dest or self.folder or ""
        return self.run_command(
            self.init_cmd(dest)
            + [
                "-c",
                "commit.gpgsign=false",
                "commit",
                "-a",
                "-m",
                message,
                "--allow-empty",
            ]
        )

    def status(self, dest=None):
        """
        Add a file to the git repository
        """
        dest = dest or self.folder or ""
        return self.run_command(self.init_cmd(dest) + ["status"])

    def clone(self, repo, dest=None):
        """
        Given a repository, clone it with run_command
        """
        # Destination folder can default to present working directory
        dest = dest or self.folder or ""
        self.run_command(["git", "clone", "--depth", "1", repo, dest])

        # Disable warnings about detached head
        self.config("advice.detachedHead", "false", dest)
        self.update_repo(dest)
        return dest

    def init(self, dest=None):
        """
        Init an empty repository in a directory of choice
        """
        dest = dest or self.folder or ""
        self.run_command(["git", "init", dest])
        self.update_repo(dest)
        self.config("user.name", "contributor-ci", dest)
        self.config("user.email", "contributor-ci@users.noreply.github.com", dest)
        return dest

    def config(self, key, value, dest=None):
        self.run_command(self.init_cmd(dest) + ["config", key, value])

    def checkout(self, commit, dest=None):
        self.run_command(self.init_cmd(dest) + ["checkout", commit])

    def ls_files(self, dest=None):
        """
        Init an empty repository in a directory of choice
        """
        dest = dest or self.folder or ""
        files = self.run_command(self.init_cmd(dest) + ["ls-files"]) or []
        if files:
            return [x for x in files[0].split("\n") if x]
        return files

    def tag(self, tag, dest=None):
        """
        Create a tag for a particular commit
        """
        dest = dest or self.folder or ""
        return self.run_command(self.init_cmd(dest) + ["tag", tag])

    def init_cmd(self, dest):
        """
        Initialize a command that sets the git-dir and working tree
        """
        dest = dest or "."
        git_dir = os.path.join(dest, ".git")
        return [
            "git",
            "--git-dir=%s" % git_dir,
            "--work-tree=%s" % os.path.dirname(git_dir),
        ]

    def run_command(self, cmd):
        """
        A wrapper to run_command to handle errors
        """
        logger.debug(" ".join(cmd))
        response = run_command(cmd)
        if not response["return_code"] == 0:
            logger.exit("Error with %s, %s" % (" ".join(cmd), response["message"]))
        return response["message"]



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"


import os
from datetime import datetime

import jsonschema

import contributor_ci.defaults as defaults
import contributor_ci.main.schemas
import contributor_ci.utils
from contributor_ci.logger import logger


class SettingsBase:
    def __init__(self):
        """
        Create a new settings object not requiring a settings file.
        """
        # Set an updated time, in case it's written back to file
        self._settings = {"updated_at": datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")}
        self.settings_file = None

    def __str__(self):
        return "[contributor-ci-settings]"

    def __repr__(self):
        return self.__str__()

    def validate(self):
        """
        Validate the loaded settings with jsonschema
        """
        jsonschema.validate(
            instance=self._settings, schema=contributor_ci.main.schemas.config_schema
        )

    def load(self, settings_file=None):
        """
        Load the settings file into the settings object
        """
        self.settings_file = settings_file or defaults.default_settings_file

        # Exit quickly if the settings file does not exist
        if not os.path.exists(self.settings_file):
            logger.exit("%s does not exist." % self.settings_file)

        # Load the yaml file into settings
        self._settings = contributor_ci.utils.read_yaml(self.settings_file)

    def add(self, key, value):
        """
        Given an existing key and a value, add the value to the list.
        """
        current = self.get(key)
        if not current:
            logger.exit("%s is not a known key to add to." % key)
        if not isinstance(current, list):
            logger.exit("You can only remove from a setting key that is a list.")
        logger.info("Adding %s to %s" % (value, key))
        current.append(value)
        current = sorted(list(set(current)))
        self.set(key, current)

    def remove(self, key, value):
        """
        Given an existing key and a value, remove the value.
        """
        current = self.get(key)
        if not current:
            logger.exit("%s is not a known key to remove from." % key)
        if not isinstance(current, list):
            logger.exit("You can only remove from a setting key that is a list.")
        current = set(current)
        if value not in current:
            logger.exit("%s is not in %s." % (value, key))
        logger.info("Removing %s from %s" % (value, key))
        current.remove(value)
        current = sorted(list(current))
        self.set(key, current)

    def sort(self):
        """
        Sort all lists.
        """
        for key, value in self._settings.items():
            if isinstance(value, list):
                value.sort()

    def get(self, key, default=None):
        value = self._settings.get(key, default)
        return self._substitutions(value)

    def __getattr__(self, key):
        """
        A direct get of an attribute, but default to None if doesn't exist
        """
        return self.get(key)

    def edit(self):
        """
        Interactively edit a config file.
        """
        editor = self.editor or "vim"
        if not self.settings_file or not os.path.exists(self.settings_file):
            logger.exit("Settings file not found.")
        contributor_ci.utils.run_command([editor, self.settings_file], stream=True)

    def set(self, key, value):
        """
        Set a setting based on key and value. If the key has :, it's nested
        """
        value = True if value == "true" else value
        value = False if value == "false" else value

        # This is a reference to a dictionary (object) setting
        if ":" in key:
            key, subkey = key.split(":")
            self._settings[key][subkey] = value
        else:
            self._settings[key] = value

    def _substitutions(self, value):
        """
        Given a value, make substitutions
        """
        if isinstance(value, bool) or not value:
            return value

        # Currently dicts only support boolean or null so we return as is
        elif isinstance(value, dict):
            return value

        for rep, repvalue in defaults.reps.items():
            if isinstance(value, list):
                value = [x.replace(rep, repvalue) for x in value]
            else:
                value = value.replace(rep, repvalue)

        return value

    def delete(self, key):
        if key in self._settings:
            del self._settings[key]

    def save(self, filename=None):
        filename = filename or self.settings_file
        if not filename:
            logger.exit("A filename is required to save to.")
        contributor_ci.utils.write_yaml(self._settings, filename)

    def __iter__(self):
        for key, value in self.__dict__.items():
            yield key, value


class Settings(SettingsBase):
    """
    The settings class is a wrapper for easily parsing a settings.yml file.

    We parse into a query-able class. It also gives us control to update settings,
    meaning we change the values and then write them to file. It's basically
    a dictionary-like class with extra functions.
    """

    def __init__(self, settings_file):
        """
        Create a new settings object, which requires a settings file to load
        """
        self.load(settings_file)
        self.validate()



__author__ = "Vanessa Sochat"
__copyright__ = "Copyright 2021, Vanessa Sochat"
__license__ = "MPL 2.0"

from contributor_ci.logger import logger
from contributor_ci.main.extractor import GitHubExtractorBase


class Dependencies(GitHubExtractorBase):

    name = "dependencies"
    description = "extract dependencies."
    filenames = ["dependencies"]
    depends_on = ["repo_dependencies"]

    def extract(self):
        """
        Extract more metadata about repository dependencies.
        """
        dep_query = self.get_local_query("dependencies-info.gql")

        # This extract only saves one result - a repos file
        self._data[self.name] = {}

        # We require internal repositories list as input
        depends = self.load_dependency_file("repo_dependencies")
        logger.info("Found %s dependencies to find metadata for." % len(depends.data))

        repolist = []
        # Generate a list (will have duplicates) of dependency names
        for i, name in enumerate(depends.data):
            for node in depends.data[name]["dependencyGraphManifests"]["nodes"]:
                for repo in node["dependencies"]["nodes"]:
                    if repo["repository"] and repo["repository"]["nameWithOwner"]:
                        repolist.append(repo["repository"]["nameWithOwner"])

        # This ensures unique!
        repolist = list(dict.fromkeys(repolist))
        repolist.sort()
        logger.info("Found %s dependency repositories." % len(repolist))
        for i, repo in enumerate(repolist):
            repo_user, repo_name = repo.split("/")
            logger.info(
                "Getting dependency metadata for %s, %s of %s"
                % (repo, i, len(repolist))
            )
            try:
                out = self.manager.queryGitHubFromFile(
                    dep_query,
                    {"ownName": repo_user, "repoName": repo_name},
                    headers={"Accept": "application/vnd.github.hawkgirl-preview+json"},
                )
            except Exception as error:
                logger.warning("Warning: Could not complete '%s': %s" % (repo, error))
                continue

            self._data[self.name][repo] = out["data"]["repository"]
